/*
 * JBoss, Home of Professional Open Source
 * Copyright 2006, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags. 
 * See the copyright.txt in the distribution for a full listing 
 * of individual contributors.
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions
 * of the GNU Lesser General Public License, v. 2.1.
 * This program is distributed in the hope that it will be useful, but WITHOUT A
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License,
 * v.2.1 along with this distribution; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 * 
 * (C) 2005-2006,
 * @author JBoss Inc.
 */
/*
 * Copyright (C) 2002,
 *
 * Arjuna Technologies Limited,
 * Newcastle upon Tyne,
 * Tyne and Wear,
 * UK.
 *
 * $Id: CoordinatorManager.java,v 1.1 2003/01/07 10:33:41 nmcl Exp $
 */

package com.arjuna.mw.wscf.model.as;

import com.arjuna.mw.wscf.common.*;

import com.arjuna.mw.wscf.model.as.coordinator.*;

import com.arjuna.mw.wsas.activity.Outcome;

import com.arjuna.mw.wscf.exceptions.*;

import com.arjuna.mw.wsas.exceptions.WrongStateException;
import com.arjuna.mw.wsas.exceptions.SystemException;
import com.arjuna.mw.wsas.exceptions.NoActivityException;

/**
 * The CoordinatorManager is the way in which services can enlist
 * participants with the current coordinator.
 *
 * @author Mark Little (mark.little@arjuna.com)
 * @version $Id: CoordinatorManager.java,v 1.1 2003/01/07 10:33:41 nmcl Exp $
 * @since 1.0.
 */

public interface CoordinatorManager
{

    /**
     * Enrol the specified participant with the coordinator associated with
     * the current thread. If the coordinator supports a priority ordering
     * of participants, then that ordering can also be specified. Any
     * qualifiers that are to be associated with the participant are also
     * provided
     *
     * @param act The participant.
     * @param priority The priority to associate with the participant in
     * the coordinator's list.
     * @param quals Any qualifiers to be associated with the
     * participant.
     *
     * @exception NoActivityException Thrown if there is no activity associated
     * with the current thread.
     * @exception WrongStateException Thrown if the coordinator is not in a
     * state that allows participants to be enrolled.
     * @exception DuplicateParticipantException Thrown if the participant has
     * already been enrolled and the coordination protocol does not support
     * multiple entries.
     * @exception InvalidParticipantException Thrown if the participant is invalid.
     * @exception SystemException Thrown if any other error occurs.
     */

    public void addParticipant (Participant act, int priority, Qualifier[] quals) throws NoActivityException, WrongStateException, DuplicateParticipantException, InvalidParticipantException, SystemException;

    /**
     * Remove the specified participant from the coordinator associated with
     * the current thread.
     *
     * @param act The participant to remove.
     *
     * @exception NoActivityException Thrown if there is no activity associated
     * with the current thread.
     * @exception WrongStateException Thrown if the coordinator is not in a
     * state that allows participants to be removed.
     * @exception InvalidParticipantException Thrown if the participant is invalid.
     * @exception SystemException Thrown if any other error occurs.
     */

    public void removeParticipant (Participant act) throws NoActivityException, InvalidParticipantException, WrongStateException, SystemException;

    /**
     * Some coordination protocol messages may have asynchronous responses
     * or it may be possible for participants to autonomously generate
     * responses to messages that have not yet been producted by the
     * coordinator. As such, this method allows a response from a participant
     * to be passed to the coordinator. In order to ensure that the protocol
     * remains valid, it is necessary for the participant to specify what
     * message produced the response: if the response was autonomously
     * generated by the participant on the assumption it would receive this
     * message from the coordinator and the coordinator subsequently decides
     * not to produce such a message, then the action taken by the participant
     * is invalid and hence so is the response.
     *
     * @param id the unique participant identification.
     * @param notification the message the participant got/assumed
     * when producing the response.
     * @param response the actual response.
     * @param quals any qualifiers associated with the response.
     *
     * @exception InvalidParticipantException Thrown if the coordinator has no
     * knowledge of the participant.
     * @exception WrongStateException Thrown if the coordinator is in a state
     * that does not allow it to accept responses at all or this specific
     * type of response.
     * @exception SystemException Thrown if any other error occurs.
     */

    public void setResponse (String id, Message notification, Outcome response, Qualifier[] quals) throws InvalidParticipantException, NoCoordinatorException, WrongStateException, SystemException;
    
}

